import requests

from aerospike_rest import NAME, get_version
from aerospike_rest.exceptions import AerospikeRestApiError

import logging
logger = logging.getLogger(__name__)


class AerospikeRestApi(object):
    def __init__(self, base_url, authorization=None, user_agent=None,
                 connect_timeout=2, client_compression=True,
                 http_compression=True):
        """
        Requests wrapper to make API calls to Aerospike REST Client.

        Args:
            base_url            - Absolute URL to Aerospike REST Client
            authorization       - Aerospike user:password
            user_agent          - User-agent string included in requests
            connect_timeout     - Connect timeout for all requests in seconds
            client_compression  - Enable compression between REST and Aerospike
            http_compression    - Enable compression between this and REST
        """
        self.base_url = base_url.rstrip("/")
        if base_url[0:4] == "http":
            self.base_url = self.base_url
        else:
            self.base_url = "https://" + self.base_url

        self.authorization = authorization
        self.client_compression = client_compression
        self.http_compression = http_compression
        self.connect_timeout = connect_timeout

        if not user_agent:
            self.user_agent = "{}/{}".format(NAME, get_version())
        else:
            self.user_agent = user_agent

        # use keep-alive
        self.session = requests.Session()


    def request(self, request_func, path, body=None, params=None, 
                 headers=None, timeout=30):
        """
        Send HTTP request to Aerospike REST client endpoint.
        """
        all_params = {
            'compress': self.client_compression
        }

        if params:
            all_params.update(params)

        all_headers = {
            'User-Agent': self.user_agent
        }

        if self.http_compression:
            all_headers['Accept-Encoding'] = 'gzip'
        
        if self.authorization:
            all_headers['Authorization'] = self.authorization

        if headers:
            all_headers.update(headers)

        url = "{}/{}".format(self.base_url, path.lstrip("/"))
        response = request_func(url, params=all_params, json=body,
                                headers=all_headers,
                                timeout=(self.connect_timeout, timeout))

        logger.debug("--- [REQUEST] ----------------------------------")
        for k, v in response.request.headers.items():
            logger.debug("{:24}: {}".format(k, v))
        logger.debug(response.request.body)

        logger.debug("--- [RESPONSE] ---------------------------------")
        for k, v in response.headers.items():
            logger.debug("{:24}: {}".format(k, v))
        content = response.content.decode("utf-8")
        logger.debug("Body:\n{}".format(content or "<EMPTY BODY>"))

        json = self.json_from_response(response)

        return json


    def get(self, path, body=None, params=None, headers=None, timeout=30):
        """
        Return UTF-8 response body from HTTP GET request to Aerospike REST
        Client endpoint.

        Args:
            path    - API path relative to base_url
            body    - HTTP request body (sent as application/json)
            params  - Dictionary of request query string parameters
            timeout - Read timeout in seconds

        Raises:
            AerospikeRestClientError
        """
        return self.request(self.session.get, path, body, params, headers,
                            timeout)


    def post(self, path, body=None, params=None, headers=None, timeout=30):
        """
        Return UTF-8 response body from HTTP POST request to Aerospike REST
        Client endpoint.

        Args:
            path    - API path relative to base_url
            body    - HTTP request body (sent as application/json)
            params  - Dictionary of request query string parameters
            timeout - Read timeout in seconds

        Raises:
            AerospikeRestClientError
        """
        return self.request(self.session.post, path, body, params, headers,
                            timeout)


    def put(self, path, body=None, params=None, headers=None, timeout=30):
        """
        Return UTF-8 response body from HTTP PUT request to Aerospike REST
        Client endpoint.

        Args:
            path    - API path relative to base_url
            body    - HTTP request body (sent as application/json)
            params  - Dictionary of request query string parameters
            timeout - Read timeout in seconds

        Raises:
            AerospikeRestClientError
        """
        return self.request(self.session.put, path, body, params, headers,
                            timeout)


    def delete(self, path, body=None, params=None, headers=None, timeout=30):
        """
        Return UTF-8 response body from HTTP DELETE request to Aerospike REST
        Client endpoint.

        Args:
            path    - API path relative to base_url
            body    - HTTP request body (sent as application/json)
            params  - Dictionary of request query string parameters
            timeout - Read timeout in seconds

        Raises:
            AerospikeRestClientError
        """
        return self.request(self.session.delete, path, body, params, headers,
                            timeout)


    def patch(self, path, body=None, params=None, headers=None, timeout=30):
        """
        Return UTF-8 response body from HTTP PATCH request to Aerospike REST
        Client endpoint.

        Args:
            path    - API path relative to base_url
            body    - HTTP request body (sent as application/json)
            params  - Dictionary of request query string parameters
            timeout - Read timeout in seconds

        Raises:
            AerospikeRestClientError
        """
        return self.request(self.session.patch, path, body, params, headers,
                            timeout)

    def json_from_response(self, response):
        try:
            json = response.json()
            if response.status_code >= 400:
                raise AerospikeRestApiError(json, response.status_code)
            return json
        except ValueError:
            # The response was not valid JSON (empty body, 5xx errors, etc.)
            response.raise_for_status()



from unittest import TestCase
from unittest.mock import MagicMock, patch

from copy import deepcopy
from requests.exceptions import HTTPError
from json.decoder import JSONDecodeError

from aerospike_rest import NAME, get_version
from aerospike_rest.api import AerospikeRestApi
from aerospike_rest.exceptions import AerospikeRestApiError


class TestAerospikeRestApi(TestCase):
    wrapper_methods = (
        ('requests.Session.get', 'get'),
        ('requests.Session.post', 'post'),
        ('requests.Session.put', 'put'),
        ('requests.Session.patch', 'patch'),
        ('requests.Session.delete', 'delete')
    )

    default_headers = {
        'User-Agent': "{}/{}".format(NAME, get_version()),
        'Accept-Encoding': 'gzip'
    }

    default_params = {
        'compress': True
    }

    default_timeout = 30

    def test_init(self):
        """
        AerospikeRestApi initializes class properties from initializer.
        """
        kwargs = {
            'base_url': 'http://foo',
            'authorization': 'b',
            'user_agent': 'c',
            'connect_timeout': 'd',
            'client_compression': 'e',
            'http_compression': 'f',
        }

        api = AerospikeRestApi(**kwargs)
        for k, v in kwargs.items():
            self.assertEqual(getattr(api, k), v)

    def test_url_sanitization(self):
        """
        AerospikeRestApi enforces scheme and strips trailing slash.
        """
        expected_url = 'https://foo'

        api = AerospikeRestApi('https://foo/')
        self.assertEqual(api.base_url, expected_url)

        api = AerospikeRestApi('foo/')
        self.assertEqual(api.base_url, expected_url)

    def test_method_wrappers(self):
        """
        AerospikeRestApi HTTP method wrappers map to requests.session functions.
        """
        api = AerospikeRestApi('http://foo')

        for patch_method, api_method in self.wrapper_methods:
            with patch(patch_method) as mock:
                mock.return_value.status_code = 200
                getattr(api, api_method)('/bar')
                mock.assert_called_once()

    def _mock_client_success_response(self):
        return {
            "inDoubt": False,
            "internalErrorCode": 0,
            "message": "success message"
        }

    def test_success_response(self):
        """
        AerospikeRestApi returns JSON on success.
        """
        for patch_method, api_method in self.wrapper_methods:
            with patch(patch_method) as mock:
                api = AerospikeRestApi('http://foo')
                obj = MagicMock()
                obj.status_code = 200
                obj.json = self._mock_client_success_response
                mock.return_value = obj
                json = getattr(api, api_method)("/bar")
                self.assertEqual(json, self._mock_client_success_response())

    def _mock_client_error_response(self):
        return {
            "inDoubt": True,
            "internalErrorCode": -1,
            "message": "error message"
        }

    def test_expected_status_exceptions(self):
        """
        AerospikeRestApi raises AerospikeRestApiError for expected HTTP status
        codes.
        """
        with patch('requests.Session.get') as mock:
            for status_code in (403, 404, 409):
                api = AerospikeRestApi('http://foo')
                obj = MagicMock()
                obj.status_code = status_code
                obj.json = self._mock_client_error_response
                mock.return_value = obj
                with self.assertRaises(AerospikeRestApiError) as cm:
                    api.get("/bar")
                self.assertEqual(cm.exception.status_code, status_code)

    def _mock_json_decode_error(self):
        raise JSONDecodeError('test message', 'doc', 1)

    def test_unexpected_status_exceptions(self):
        """
        AerospikeRestApi raises underlying exception for unexpected status
        codes.
        """
        with patch('requests.Session.get') as mock:
            for status_code in (410, 205, 500):
                api = AerospikeRestApi('http://foo')
                obj = MagicMock()
                obj.status_code = status_code
                obj.json = self._mock_json_decode_error
                mock.return_value = obj
                api.get("/bar")
                obj.raise_for_status.assert_called()

    def _success_called_with(self, api, headers=None, params=None,
                             timeout=30):
        """
        Helper method to test successful requests get called a certain way
        """
        for patch_method, api_method in self.wrapper_methods:
            with patch(patch_method) as mock:
                obj = MagicMock()
                obj.status_code = 200
                obj.json = self._mock_client_success_response
                mock.return_value = obj

                getattr(api, api_method)("/bar", headers=headers, params=params,
                                         timeout=timeout)

                all_headers = deepcopy(self.default_headers)
                all_headers['User-Agent'] = api.user_agent
                if headers:
                    all_headers.update(headers)
                if not api.http_compression:
                    all_headers.pop('Accept-Encoding')
                if api.authorization:
                    all_headers['Authorization'] = api.authorization

                all_params = deepcopy(self.default_params)
                if params:
                    all_params.update(params)
                if not api.client_compression:
                    all_params['compress'] = False

                if not timeout:
                    timeout = self.default_timeout

                mock.assert_called_with("http://foo/bar", headers=all_headers,
                                        params=all_params, json=None,
                                        timeout=(api.connect_timeout, timeout))

    def test_optional_headers(self):
        """
        AerospikeRestApi should override default headers with arg headers
        """
        api = AerospikeRestApi('http://foo')
        self._success_called_with(api, headers={"Foo": "Bar"})

    def test_optional_params(self):
        """
        AerospikeRestApi should override default params with arg params
        """
        api = AerospikeRestApi('http://foo')
        self._success_called_with(api, params={"Foo": "Bar"})

    def test_optional_timeout(self):
        """
        AerospikeRestApi should override default timeouts
        """
        api = AerospikeRestApi('http://foo')
        api.connect_timeout = 999
        self._success_called_with(api, timeout=999)

    def test_optional_compression(self):
        """
        AerospikeRestApi should allow override of default compression
        """
        api = AerospikeRestApi('http://foo')
        api.client_compression = False
        api.http_compression = False
        self._success_called_with(api)

    def test_optional_authorization(self):
        """
        AerospikeRestApi should allow optional authorization header
        """
        api = AerospikeRestApi('http://foo')
        api.authorization = 'Basic abcdefg='
        self._success_called_with(api)

    def test_optional_user_agent(self):
        """
        AerospikeRestApi should allow optional user-agent
        """
        api = AerospikeRestApi('http://foo')
        api.user_agent = 'bar'
        self._success_called_with(api)

    def test_json_from_response_valid(self):
        mock_response = MagicMock()
        expected_json = "\"valid json\""
        attrs = {'json.return_value': expected_json, 'status_code': 200}
        mock_response.configure_mock(**attrs)

        api = AerospikeRestApi('http://foo')
        actual_json = api.json_from_response(mock_response)

        self.assertEqual(actual_json, expected_json)

    def test_json_from_response_400_error(self):
        mock_response = MagicMock()
        expected_json = {'message': 'valid json', 'inDoubt': 'valid json', 'internalErrorCode': 'valid json'}
        attrs = {'json.return_value': expected_json, 'status_code': 400}
        mock_response.configure_mock(**attrs)

        api = AerospikeRestApi('http://foo')
        with self.assertRaises(AerospikeRestApiError):
            api.json_from_response(mock_response)

    def test_json_from_response_no_json_value_error(self):
        mock_response = MagicMock()
        attrs = {'json.side_effect': ValueError, 'raise_for_status.side_effect': HTTPError, 'status_code': 500}
        mock_response.configure_mock(**attrs)

        api = AerospikeRestApi('http://foo')
        with self.assertRaises(HTTPError):
            api.json_from_response(mock_response)



#!/usr/bin/env python

import os
from datetime import datetime
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

import aerospike_rest

version = '{}{}'.format(
    aerospike_rest.get_version(),
    'b{}'.format(datetime.now().strftime("%Y%m%d%H%M%S")) if os.getenv('TEST_PYPI', '') else ''
)

setup(
    name = aerospike_rest.NAME,
    version = version,
    description = "Python interface to Aerospike REST Client",
    url = "https://github.com/aerospike-community/aerospike-python-rest",
    long_description = read("README.md"),
    long_description_content_type='text/markdown',
    packages = find_packages(),
    python_requires=">=3.3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        'requests'
    ],
)




class AerospikeRestApiError(Exception):
    def __init__(self, http_response, status_code):
        self.message = http_response['message']
        self.in_doubt = http_response['inDoubt']
        self.status_code = status_code
        self.code = http_response['internalErrorCode']
    
    def __str__(self):
        return "{} {}".format(self.status_code, self.message)


# error codes from Java client:
# https://github.com/aerospike/aerospike-client-java/blob/master/client/src/com/aerospike/client/ResultCode.java

SERIALIZE_ERROR = -10
ASYNC_QUEUE_FULL = -9
SERVER_NOT_AVAILABLE = -8
NO_MORE_CONNECTIONS = -7
COMMAND_REJECTED = -6
QUERY_TERMINATED = -5
SCAN_TERMINATED = -4
INVALID_NODE_ERROR = -3
PARSE_ERROR = -2
CLIENT_ERROR = -1
OK = 0
SERVER_ERROR = 1
KEY_NOT_FOUND_ERROR = 2
GENERATION_ERROR = 3
PARAMETER_ERROR = 4
KEY_EXISTS_ERROR = 5
BIN_EXISTS_ERROR = 6
CLUSTER_KEY_MISMATCH = 7
SERVER_MEM_ERROR = 8
TIMEOUT = 9
ALWAYS_FORBIDDEN = 10
PARTITION_UNAVAILABLE = 11
BIN_TYPE_ERROR = 12
RECORD_TOO_BIG = 13
KEY_BUSY = 14
SCAN_ABORT = 15
UNSUPPORTED_FEATURE = 16
BIN_NOT_FOUND = 17
DEVICE_OVERLOAD = 18
KEY_MISMATCH = 19
INVALID_NAMESPACE = 20
BIN_NAME_TOO_LONG = 21
FAIL_FORBIDDEN = 22
ELEMENT_NOT_FOUND = 23
ELEMENT_EXISTS = 24
ENTERPRISE_ONLY = 25
OP_NOT_APPLICABLE = 26
FILTERED_OUT = 27
QUERY_END = 50
SECURITY_NOT_SUPPORTED = 51
SECURITY_NOT_ENABLED = 52
SECURITY_SCHEME_NOT_SUPPORTED = 53
INVALID_COMMAND = 54
INVALID_FIELD = 55
ILLEGAL_STATE = 56
INVALID_USER = 60
USER_ALREADY_EXISTS = 61
INVALID_PASSWORD = 62
EXPIRED_PASSWORD = 63
FORBIDDEN_PASSWORD = 64
INVALID_CREDENTIAL = 65
EXPIRED_SESSION = 66
INVALID_ROLE = 70
ROLE_ALREADY_EXISTS = 71
INVALID_PRIVILEGE = 72
INVALID_WHITELIST = 73
NOT_AUTHENTICATED = 80
ROLE_VIOLATION = 81
NOT_WHITELISTED = 82
UDF_BAD_RESPONSE = 100
BATCH_DISABLED = 150
BATCH_MAX_REQUESTS_EXCEEDED = 151
BATCH_QUEUES_FULL = 152
INDEX_ALREADY_EXISTS = 200
INDEX_FOUND = 200
INDEX_NOTFOUND = 201
INDEX_OOM = 202
INDEX_NOTREADABLE = 203
INDEX_GENERIC = 204
INDEX_NAME_MAXLEN = 205
INDEX_MAXCOUNT = 206
QUERY_ABORTED = 210
QUERY_QUEUEFULL = 211
QUERY_TIMEOUT = 212
QUERY_GENERIC = 213
